#include <syslog.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#ifdef HAVE_LIBCRYPT
	#include <crypt.h>
#endif
#include <grp.h>

#include "config.h"
#include "globals.h"
#include "database.hpp"
#include "obscure.h"
#include "server.hpp"
#include "sockio.h"
#include "salt.h"
#include "conf.hpp"

extern CConfigurationFile *pConfigFile;

void server(int sockfd)
{
	char szLogin[32];
	char szPassword[32];
	
	write_byte(sockfd, WELCOME);
	if (-1 == read_string(sockfd, szLogin, sizeof(szLogin)))
		write_byte(sockfd, TIMEOUT);
	if (-1 == read_string(sockfd, szPassword, sizeof(szPassword)))
		write_byte(sockfd, TIMEOUT);
	if (false == check_login(szLogin, szPassword)) {
		write_byte(sockfd, LOGIN_FAILED);
		return;
	}
	write_byte(sockfd, GO_AHEAD);
	
	unsigned char cAction = '\0';
	if (true != read_byte(sockfd, &cAction))
		write_byte(sockfd, TIMEOUT);
	
	switch (cAction) {
		case ACTION_NOTHING:
			syslog(LOG_USER | LOG_INFO, "user %s did nothing", szLogin);
		break;
		case ACTION_CHANGEPASSWORD:
			{
				char szTarget[32];
				char szNewPassword[32];
				int iMin, iMax, iWarn, iInact;
				
				if ((-1 == read_string(sockfd, szTarget, sizeof(szTarget))) ||
					 (-1 == read_string(sockfd, szNewPassword, sizeof(szNewPassword))) ||
					 (false == read_integer(sockfd, &iMin)) ||
					 (false == read_integer(sockfd, &iMax)) ||
					 (false == read_integer(sockfd, &iWarn)) ||
					 (false == read_integer(sockfd, &iInact))) {
					write_byte(sockfd, TIMEOUT);
					return;
				}
				
				syslog(LOG_USER | LOG_INFO, "user %s tries to change password of %s", szLogin, szTarget);
				if (false == check_target(szTarget)) {
					write_byte(sockfd, INVALID_USER);
					return;
				}
				if (false == allow_change(szLogin, szTarget)) {
					write_byte(sockfd, NO_RIGHTS);
					return;
				}
				char *pszObscure = obscure(szPassword, szNewPassword, 0 == strcmp(szLogin, szTarget));
				if (NULL != pszObscure) {
					write_byte(sockfd, WEAK_PASSWORD);
					write_string(sockfd, pszObscure);
					return;
				}
				uid_t i = getuid();
				if (false == change_password(szTarget, szNewPassword, iMin, iMax,	iWarn, iInact)) {
					write_byte(sockfd, ERROR);
					setuid(i);
					return;
				}
				write_byte(sockfd, OK);
				setuid(i);
				return;
			}
		break;
		case ACTION_ADDUSER:
			{
				char szTarget[32];
				char szGecos[128];
				
				if ((-1 == read_string(sockfd, szTarget, sizeof(szTarget))) ||
					 (-1 == read_string(sockfd, szGecos, sizeof(szGecos)))) {
					write_byte(sockfd, TIMEOUT);
					return;
				}
				
				int j, k = 0;
				for (j = 0; '\0' != szGecos[j]; j++) {
					if (',' == szGecos[j])
						k += 1;
					if (k > 4 ||
						 ';' == szGecos[j] ||
						 '=' == szGecos[j] ||
						 iscntrl(szGecos[j])) {
						write_byte(sockfd, ERROR);
						return;
					}
				}
				if ((j > 0) &&
					 (0 == strcmp(szLogin, szTarget)) &&
					 (',' != szGecos[strlen(szGecos) - 1])) {
					write_byte(sockfd, ERROR);
					return;
				}
				
				syslog(LOG_USER | LOG_INFO, "user %s tries to add user %s", szLogin, szTarget);
				
				if (true == check_target(szTarget)) {
					write_byte(sockfd, INVALID_USER);
					return;
				}
				if (false == allow_add(szLogin)) {
					write_byte(sockfd, NO_RIGHTS);
					return;
				}
				uid_t i = getuid();
				if (false == add_user(szTarget, szGecos)) {
					write_byte(sockfd, ERROR);
					setuid(i);
					return;
				}
				write_byte(sockfd, OK);
				setuid(i);
				return;
			}
		break;
		case ACTION_DELETEUSER:
			{
				char szTarget[32];
				
				if (-1 == read_string(sockfd, szTarget, sizeof(szTarget))) {
					write_byte(sockfd, TIMEOUT);
					return;
				}
				
				syslog(LOG_USER | LOG_INFO, "user %s tries to delete user %s", szLogin, szTarget);
				
				if (false == check_target(szTarget)) {
					write_byte(sockfd, INVALID_USER);
					return;
				}
				if (false == allow_change(szLogin, szTarget)) {
					write_byte(sockfd, NO_RIGHTS);
					return;
				}
				uid_t i = getuid();
				if (false == delete_user(szTarget)) {
					write_byte(sockfd, ERROR);
					setuid(i);
					return;
				}
				write_byte(sockfd, OK);
				setuid(i);
				return;
			}
		break;
		case ACTION_GETGROUP:
			{
				char szTarget[32];
				
				if (-1 == read_string(sockfd, szTarget, sizeof(szTarget))) {
					write_byte(sockfd, TIMEOUT);
					return;
				}
				
				syslog(LOG_USER | LOG_INFO, "user %s tries to get group name of user %s", szLogin, szTarget);
				
				if (false == check_target(szTarget)) {
					write_byte(sockfd, INVALID_USER);
					return;
				}
				if (false == allow_change(szLogin, szTarget)) {
					write_byte(sockfd, NO_RIGHTS);
					return;
				}
				char *pszGroup = get_group(szTarget);
				write_string(sockfd, pszGroup);
				if ('\0' == *pszGroup) {
					write_byte(sockfd, ERROR);
					return;
				}
				write_byte(sockfd, OK);
				return;
			}
		break;
		default:
			write_byte(sockfd, INVALID_ACTION);
		break;
	}
	
	write_byte(sockfd, GOOD_BYE);
}

bool allow_change(char *pszLogin, char *pszTarget)
{
	static char szBuf[256];
	static char szTmp[256];
	const struct group *grLogin;
	const struct group *grTarget;
	char *pszAllowed;
	char *pszPowered;
	char *pszSheep;
	bool fgAllowed = false;
	bool fgPowered = false;
	bool fgSheep = false;
	
	CPasswdDatabase cPasswdDatabase(true);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");		
		return false;
	}
	gid_t iLoginGroup = cPasswdDatabase.GetGroup(pszLogin);
	gid_t iTargetGroup = cPasswdDatabase.GetGroup(pszTarget);
	
	char *pszAllowedUsers = pConfigFile->GetString("Permissions", "users_allow", "");
	char *pszPoweredUsers = pConfigFile->GetString("Permissions", "master", "");
	
	register int i, j;
	for (i = j = 0; pszAllowedUsers[i] != '\0'; i++)
	   if (NULL == strchr(" \t\x0a\x0d", pszAllowedUsers[i]))
		   szBuf[j++] = pszAllowedUsers[i];
	szBuf[j] = '\0';
	for (i = j = 0; pszPoweredUsers[i] != '\0'; i++)
		if (NULL == strchr(" \t\x0a\x0d", pszPoweredUsers[i]))
			szTmp[j++] = pszPoweredUsers[i];
	szTmp[j] = '\0';
	if (NULL == (grLogin = getgrgid(iLoginGroup))) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszLogin);
		return false;
	}
	if (NULL == (grTarget = getgrgid(iTargetGroup))) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszTarget);
		return false;
	}
	if (NULL != (pszAllowed = strtok(szBuf, ",;")))
		do {
			if ('@' == pszAllowed[0]) {
			  	if (0 == strcmp(pszAllowed + 1, grTarget->gr_name)) {
					fgAllowed = true;
					break;
				}
			} else {
			   if (0 == strcmp(pszLogin, pszAllowed)) {
					fgAllowed = true;
					break;
				}
			}
		} while (NULL != (pszAllowed = strtok(NULL, ",;")));
	if (false == fgAllowed) {
		syslog(LOG_USER | LOG_INFO, "unauthorized user %s", pszLogin);
		return false;
	}
	if (0 == strcmp(pszLogin, pszTarget))
		return true;
	
	if (NULL != (pszPowered = strtok(szTmp, ",;")))
		do {
			pszPowered = strcpy(szBuf, pszPowered);
			if (NULL == (pszSheep = strchr(szBuf, ':'))) {
				syslog(LOG_USER | LOG_INFO, "configuration file error, skipping power users");
				continue;
			}
			*pszSheep = '\0';
			pszSheep += 1;
			if ('@' == pszPowered[0]) {
				if (0 == strcmp(pszPowered + 1, grLogin->gr_name))
					fgPowered = true;
			} else {
				if (0 == strcmp(pszPowered, pszLogin))
					fgPowered = true;
			}
			if (fgPowered && ('@' == pszSheep[0])) {
				if (0 == strcmp(pszSheep + 1, grTarget->gr_name)) {
					fgSheep = true;
					break;
				}
			} else {
				if (0 == strcmp(pszSheep, pszTarget)) {
					fgSheep = true;
					break;
				}
			}
		} while (NULL != (pszPowered = strtok(NULL, ",;")));
	if (true == fgSheep)
		return true;
	syslog(LOG_USER | LOG_INFO, "%s tried to change the password of %s without rights", pszLogin, pszTarget);
	return false;
}

bool change_password(char *pszLogin,
							char *pszNewPassword,
							int iMin,
							int iMax,
							int iWarn,
							int iInact)
{
	char *pszCryptedPassword = crypt(pszNewPassword, crypt_make_salt());
	
	memset(pszNewPassword, 0, strlen(pszNewPassword));		// Clear password

	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords setuid error: %s", strerror(errno));
		return false;
	}
	
#ifdef SHADOWPWD
	CShadowDatabase cShadowDatabase(false);
	if (NO_ERROR != cShadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading shadow file");
		return false;
	}
	if (false == cShadowDatabase.ChangePassword(pszLogin,
															  pszCryptedPassword,
															  iMin,
															  iMax,
															  iWarn,
															  iInact)) {
		syslog(LOG_USER | LOG_INFO, "error modifying shadow file");
		return false;
	}
#else
	CPasswdDatabase cPasswdDatabase(false);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	if (false == cPasswdDatabase.ChangePassword(pszLogin,
															  pszCryptedPassword,
															  iMin,
															  iMax,
															  iWarn,
															  iInact)) {
		syslog(LOG_USER | LOG_INFO, "error modifying shadow file");
		return false;
	}
#endif
	
	return true;
}

bool check_login(char *pszLogin, char *pszPassword)
{
// Check to see if there is a valid login name
	CPasswdDatabase cPasswdDatabase(true);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	if (false == cPasswdDatabase.FindUser(pszLogin)) {	
		syslog(LOG_USER | LOG_INFO, "invalid user: %s", pszLogin);
		return false;
	}
#ifdef SHADOWPWD
	uid_t i = getuid();
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords setuid error: %s", strerror(errno));
		return false;
	}												// With shadow enabled we need to be root to
													// open the file
	CShadowDatabase cShadowDatabase(true);
	if (NO_ERROR != cShadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading shadow file");
		return false;
	}
	if (false == cShadowDatabase.FindUser(pszLogin)) {
		syslog(LOG_USER | LOG_INFO, "invalid user: %s", pszLogin);
		return false;
	}
	char *pszUserPassword = cShadowDatabase.GetPassword(pszLogin);
	if ((NULL == pszUserPassword) ||
		 (0 != strcmp(pszUserPassword, crypt(pszPassword, pszUserPassword)))) {
		syslog(LOG_USER | LOG_INFO, "invalid password for user: %s", pszLogin);
		return false;
	}
#else
	char *pszUserPassword = cPasswdDatabase.GetPassword(pszLogin);
	if ((NULL == pszUserPassword) || 
		 (0 != strcmp(pszUserPassword, crypt(pszPassword, pszUserPassword)))) {
		syslog(LOG_USER | LOG_INFO, "invalid password for user: %s", pszLogin);
		return false;
	}
#endif
#ifdef SHADOWPWD
	setuid(i);									// Return to the original uid, paranoid, huh?
#endif	
	return true;
}

bool check_target(char *pszTarget)
{
// Check to see if there is a valid login name
	CPasswdDatabase cPasswdDatabase(true);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	if (false == cPasswdDatabase.FindUser(pszTarget)) {	
//		syslog(LOG_USER | LOG_INFO, "invalid user: %s", pszTarget);
		return false;
	}
#ifdef SHADOWPWD
	uid_t i = getuid();
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords setuid error: %s", strerror(errno));
		return false;
	}												// With shadow enabled we need to be root to
													// open the file
	CShadowDatabase cShadowDatabase(true);
	if (NO_ERROR != cShadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading shadow file");
		return false;
	}
	if (false == cShadowDatabase.FindUser(pszTarget)) {
//		syslog(LOG_USER | LOG_INFO, "invalid user: %s", pszTarget);
		return false;
	}
	setuid(i);
#endif	
	return true;
}

bool delete_user(char *pszTarget)
{
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords setuid error: %s", strerror(errno));
		return false;
	}
	
	CPasswdDatabase cPasswdDatabase(false);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	if (false == cPasswdDatabase.DeleteEntry(pszTarget)) {
		syslog(LOG_USER | LOG_INFO, "error modifying shadow file");
		return false;
	}

#ifdef SHADOWPWD
	CShadowDatabase cShadowDatabase(false);
	if (NO_ERROR != cShadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading shadow file");
		return false;
	}
	if (false == cShadowDatabase.DeleteEntry(pszTarget)) {
		syslog(LOG_USER | LOG_INFO, "error modifying shadow file");
		return false;
	}
#endif
	
	return true;
}

bool allow_add(char *pszLogin)
{
	static char szBuf[256];
	const struct group *grLogin;
	char *pszAllowed;
	
	CPasswdDatabase cPasswdDatabase(true);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	gid_t iLoginGroup = cPasswdDatabase.GetGroup(pszLogin);
	
	char *pszAllowedUsers = pConfigFile->GetString("Add", "master", "");
	
	register int i, j;
	for (i = j = 0; pszAllowedUsers[i] != '\0'; i++)
	   if (NULL == strchr(" \t\x0a\x0d", pszAllowedUsers[i]))
		   szBuf[j++] = pszAllowedUsers[i];
	szBuf[j] = '\0';
	if (NULL == (grLogin = getgrgid(iLoginGroup))) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszLogin);
		return false;
	}
	if (NULL != (pszAllowed = strtok(szBuf, ",;")))
		do {
			if ('@' == pszAllowed[0]) {
			  	if (0 == strcmp(pszAllowed + 1, grLogin->gr_name))
					return true;
			} else {
			   if (0 == strcmp(pszLogin, pszAllowed))
					return true;
			}
		} while (NULL != (pszAllowed = strtok(NULL, ",;")));
	return false;
}

bool add_user(char *pszTarget, char *pszGecos)
{
	char *pszDirectory = pConfigFile->GetString("Add", "directory", "/tmp");
	char *pszShell = pConfigFile->GetString("Add", "shell", "/dev/null");
	char *pszGroup = pConfigFile->GetString("Add", "group", "nogroup");
	
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords setuid error: %s", strerror(errno));
		return false;
	}
	
	CPasswdDatabase cPasswdDatabase(false);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	
	struct group *grTarget = getgrnam(pszGroup);
	if (NULL == grTarget)
		grTarget = getgrgid(atoi(pszGroup));
	if (NULL == grTarget) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszTarget);
		return false;
	}
	
	unsigned int iUserID = cPasswdDatabase.GetMaxUserID() + 1;
	
	if (false == cPasswdDatabase.AddEntry(pszTarget, iUserID, grTarget->gr_gid, pszGecos, pszDirectory, pszShell)) {
		syslog(LOG_USER | LOG_INFO, "error modifying shadow file");
		return false;
	}

#ifdef SHADOWPWD
	CShadowDatabase cShadowDatabase(false);
	if (NO_ERROR != cShadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading shadow file");
		return false;
	}
	if (false == cShadowDatabase.AddEntry(pszTarget)) {
		syslog(LOG_USER | LOG_INFO, "error modifying shadow file");
		return false;
	}
#endif
	
	return true;
}

char *get_group(char *pszTarget)
{
	CPasswdDatabase cPasswdDatabase(false);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return "";
	}
	struct group *grTarget = getgrgid(cPasswdDatabase.GetGroup(pszTarget));
	if (NULL == grTarget)
		return "";
	
	return grTarget->gr_name;
}